#include "StdAfx.h"
#include "Vector.h"

//CMatrixNX
#ifdef _DEBUG
//#define _ASSERT_ADD_X_ROTATION
//#define _ASSERT_ADD_Y_ROTATION
//#define _ASSERT_ADD_Z_ROTATION
//#define _ASSERT_ADD_TRANSLATION
//#define _ASSERT_ADD_SCALE
//#define _ASSERT_INVERSE_MATRIX
//#define _ASSERT_INVERSE_MATRIX2	// InverseMatrix()̕ʂ̎Zo@AT[g
#endif

CMatrix::CMatrix()
{
	for(int col=0; col<4; col++){
		for(int row=0; row<4; row++){
			if(row==col){
				m_rElement[row][col] = double(1.0);
			}
			else{
				m_rElement[row][col] = double(0.0);
			}
		}
	}
}

CMatrix::CMatrix(CMatrix const& mtrx)
{
	*this = mtrx;
}



#ifdef _DEBUG
double CMatrix::GetElement(int row, int col) const
{
	assert(0<=row && row<4);
	assert(0<=col && col<4);

	return m_rElement[row][col];
}
#else
//inline double CMatrix::GetElement(int row, int col) const
double CMatrix::GetElement(int row, int col) const
{
	return m_rElement[row][col];
}
#endif


#ifdef _DEBUG
CMatrix& CMatrix::SetElement(int row, int col, double val)
{
	assert(0<=row && row<4);
	assert(0<=col && col<4);

	m_rElement[row][col] = val;

	return *this;
}
#else
//inline CMatrix& CMatrix::SetElement(int row, int col, double val)
CMatrix& CMatrix::SetElement(int row, int col, double val)
{
	m_rElement[row][col] = val;

	return *this;
}
#endif


double& CMatrix::operator()(int row, int col)
{
	assert(0<=row && row<4);
	assert(0<=col && col<4);

	return m_rElement[row][col];
}

CMatrix& CMatrix::SetUnit()
{
	for(int i=0; i<4; i++){
		for(int j=0; j<4; j++){
			if(j==i){
				m_rElement[j][i] = double(1.0);
			}
			else{
				m_rElement[j][i] = double(0.0);
			}
		}
	}

	return *this;
}

BOOL CMatrix::IsUnit() const
{
	for(int i=0; i<4; i++){
		for(int j=0; j<4; j++){
			if(
				(j==i && m_rElement[j][i]!=double(1.0)) ||
				         m_rElement[j][i]!=double(0.0)){

				return FALSE;
			}
		}
	}

	return TRUE;
}

CMatrix& CMatrix::AddXRotation(double rot)
{
#if defined(_ASSERT_ADD_X_ROTATION) && defined(_DEBUG)
	CMatrix	mtrxOrg = *this;
#endif
	double	nSin = (double)sin(rot),
			nCos = (double)cos(rot);

	double	rTemp10 = m_rElement[1][0]*nCos - m_rElement[2][0]*nSin,
			rTemp20 = m_rElement[1][0]*nSin + m_rElement[2][0]*nCos,

			rTemp11 = m_rElement[1][1]*nCos - m_rElement[2][1]*nSin,
			rTemp21 = m_rElement[1][1]*nSin + m_rElement[2][1]*nCos,

			rTemp12 = m_rElement[1][2]*nCos - m_rElement[2][2]*nSin,
			rTemp22 = m_rElement[1][2]*nSin + m_rElement[2][2]*nCos,

			rTemp13 = m_rElement[1][3]*nCos - m_rElement[2][3]*nSin,
			rTemp23 = m_rElement[1][3]*nSin + m_rElement[2][3]*nCos;

	m_rElement[1][0] = rTemp10;
	m_rElement[2][0] = rTemp20;

	m_rElement[1][1] = rTemp11;
	m_rElement[2][1] = rTemp21;

	m_rElement[1][2] = rTemp12;
	m_rElement[2][2] = rTemp22;

	m_rElement[1][3] = rTemp13;
	m_rElement[2][3] = rTemp23;

#if defined(_ASSERT_ADD_X_ROTATION) && defined(_DEBUG)
	{
		CMatrix	mtrxTranslation,
				mtrxResult;
		mtrxTranslation.SetXRotation(rot);
		mtrxResult = mtrxOrg * mtrxTranslation;

		for(int i=0; i<4; i++){
			for(int j=0; j<4; j++){
				assert(m_rElement[j][i]==mtrxResult.m_rElement[j][i]);
			}
		}
	}
#endif
	
	return *this;
}

CMatrix& CMatrix::SetXRotation(double rot)
{
	double	nSin = (double)sin(rot),
			nCos = (double)cos(rot);

	m_rElement[0][0] = double(1.0);
	m_rElement[1][0] = double(0.0);
	m_rElement[2][0] = double(0.0);
	m_rElement[3][0] = double(0.0);

	m_rElement[0][1] = double(0.0);
	m_rElement[1][1] = nCos;
	m_rElement[2][1] = nSin;
	m_rElement[3][1] = double(0.0);

	m_rElement[0][2] = double(0.0);
	m_rElement[1][2] = -nSin;
	m_rElement[2][2] = nCos;
	m_rElement[3][2] = double(0.0);

	m_rElement[0][3] = double(0.0);
	m_rElement[1][3] = double(0.0);
	m_rElement[2][3] = double(0.0);
	m_rElement[3][3] = double(1.0);

	return *this;
}

CMatrix& CMatrix::AddYRotation(double rot)
{
#if defined(_ASSERT_ADD_Y_ROTATION) && defined(_DEBUG)
	CMatrix	mtrxOrg = *this;
#endif
	double	nSin = (double)sin(rot),
			nCos = (double)cos(rot);

	double	rTemp00 =  m_rElement[0][0]*nCos + m_rElement[2][0]*nSin,
			rTemp20 = -m_rElement[0][0]*nSin + m_rElement[2][0]*nCos,

			rTemp01 =  m_rElement[0][1]*nCos + m_rElement[2][1]*nSin,
			rTemp21 = -m_rElement[0][1]*nSin + m_rElement[2][1]*nCos,

			rTemp02 =  m_rElement[0][2]*nCos + m_rElement[2][2]*nSin,
			rTemp22 = -m_rElement[0][2]*nSin + m_rElement[2][2]*nCos,

			rTemp03 =  m_rElement[0][3]*nCos + m_rElement[2][3]*nSin,
			rTemp23 = -m_rElement[0][3]*nSin + m_rElement[2][3]*nCos;

	m_rElement[0][0] = rTemp00;
	m_rElement[2][0] = rTemp20;

	m_rElement[0][1] = rTemp01;
	m_rElement[2][1] = rTemp21;

	m_rElement[0][2] = rTemp02;
	m_rElement[2][2] = rTemp22;

	m_rElement[0][3] = rTemp03;
	m_rElement[2][3] = rTemp23;

#if defined(_ASSERT_ADD_Y_ROTATION) && defined(_DEBUG)
	{
		CMatrix	mtrxTranslation,
				mtrxResult;
		mtrxTranslation.SetYRotation(rot);
		mtrxResult = mtrxOrg * mtrxTranslation;

		for(int i=0; i<4; i++){
			for(int j=0; j<4; j++){
				assert(m_rElement[j][i]==mtrxResult.m_rElement[j][i]);
			}
		}
	}
#endif
	
	return *this;
}

CMatrix& CMatrix::SetYRotation(double rot)
{
	double	nSin = (double)sin(rot),
			nCos = (double)cos(rot);

	m_rElement[0][0] = nCos;
	m_rElement[1][0] = double(0.0);
	m_rElement[2][0] = -nSin;
	m_rElement[3][0] = double(0.0);

	m_rElement[0][1] = double(0.0);
	m_rElement[1][1] = double(1.0);
	m_rElement[2][1] = double(0.0);
	m_rElement[3][1] = double(0.0);

	m_rElement[0][2] = nSin;
	m_rElement[1][2] = double(0.0);
	m_rElement[2][2] = nCos;
	m_rElement[3][2] = double(0.0);

	m_rElement[0][3] = double(0.0);
	m_rElement[1][3] = double(0.0);
	m_rElement[2][3] = double(0.0);
	m_rElement[3][3] = double(1.0);

	return *this;
}

CMatrix& CMatrix::AddZRotation(double rot)
{
#if defined(_ASSERT_ADD_Z_ROTATION) && defined(_DEBUG)
	CMatrix	mtrxOrg = *this;
#endif
	double	nSin = (double)sin(rot),
			nCos = (double)cos(rot);

	double	rTemp00 = m_rElement[0][0]*nCos - m_rElement[1][0]*nSin,
			rTemp10 = m_rElement[0][0]*nSin + m_rElement[1][0]*nCos,

			rTemp01 = m_rElement[0][1]*nCos - m_rElement[1][1]*nSin,
			rTemp11 = m_rElement[0][1]*nSin + m_rElement[1][1]*nCos,

			rTemp02 = m_rElement[0][2]*nCos - m_rElement[1][2]*nSin,
			rTemp12 = m_rElement[0][2]*nSin + m_rElement[1][2]*nCos,

			rTemp03 = m_rElement[0][3]*nCos - m_rElement[1][3]*nSin,
			rTemp13 = m_rElement[0][3]*nSin + m_rElement[1][3]*nCos;

	m_rElement[0][0] = rTemp00;
	m_rElement[1][0] = rTemp10;

	m_rElement[0][1] = rTemp01;
	m_rElement[1][1] = rTemp11;

	m_rElement[0][2] = rTemp02;
	m_rElement[1][2] = rTemp12;

	m_rElement[0][3] = rTemp03;
	m_rElement[1][3] = rTemp13;

#if defined(_ASSERT_ADD_Z_ROTATION) && defined(_DEBUG)
	{
		CMatrix	mtrxTranslation,
				mtrxResult;
		mtrxTranslation.SetZRotation(rot);
		mtrxResult = mtrxOrg * mtrxTranslation;

		for(int i=0; i<4; i++){
			for(int j=0; j<4; j++){
				assert(m_rElement[j][i]==mtrxResult.m_rElement[j][i]);
			}
		}
	}
#endif
	
	return *this;
}

CMatrix& CMatrix::SetZRotation(double rot)
{
	double	nSin = (double)sin(rot),
			nCos = (double)cos(rot);

	m_rElement[0][0] = nCos;
	m_rElement[1][0] = nSin;
	m_rElement[2][0] = double(0.0);
	m_rElement[3][0] = double(0.0);

	m_rElement[0][1] = -nSin;
	m_rElement[1][1] = nCos;
	m_rElement[2][1] = double(0.0);
	m_rElement[3][1] = double(0.0);

	m_rElement[0][2] = double(0.0);
	m_rElement[1][2] = double(0.0);
	m_rElement[2][2] = double(1.0);
	m_rElement[3][2] = double(0.0);

	m_rElement[0][3] = double(0.0);
	m_rElement[1][3] = double(0.0);
	m_rElement[2][3] = double(0.0);
	m_rElement[3][3] = double(1.0);

	return *this;
}

CMatrix& CMatrix::AddTranslation(double x, double y, double z)
{
#if defined(_ASSERT_ADD_TRANSLATION) && defined(_DEBUG)
	CMatrix	mtrxOrg = *this;
#endif

	double	rTemp00 = m_rElement[0][0] + m_rElement[3][0] * x,
			rTemp10 = m_rElement[1][0] + m_rElement[3][0] * y,
			rTemp20 = m_rElement[2][0] + m_rElement[3][0] * z,
			rTemp30 = m_rElement[3][0],

			rTemp01 = m_rElement[0][1] + m_rElement[3][1] * x,
			rTemp11 = m_rElement[1][1] + m_rElement[3][1] * y,
			rTemp21 = m_rElement[2][1] + m_rElement[3][1] * z,
			rTemp31 = m_rElement[3][1],

			rTemp02 = m_rElement[0][2] + m_rElement[3][2] * x,
			rTemp12 = m_rElement[1][2] + m_rElement[3][2] * y,
			rTemp22 = m_rElement[2][2] + m_rElement[3][2] * z,
			rTemp32 = m_rElement[3][2],

			rTemp03 = m_rElement[0][3] + m_rElement[3][3] * x,
			rTemp13 = m_rElement[1][3] + m_rElement[3][3] * y,
			rTemp23 = m_rElement[2][3] + m_rElement[3][3] * z,
			rTemp33 = m_rElement[3][3];

	m_rElement[0][0] = rTemp00;
	m_rElement[1][0] = rTemp10;
	m_rElement[2][0] = rTemp20;
	m_rElement[3][0] = rTemp30;

	m_rElement[0][1] = rTemp01;
	m_rElement[1][1] = rTemp11;
	m_rElement[2][1] = rTemp21;
	m_rElement[3][1] = rTemp31;

	m_rElement[0][2] = rTemp02;
	m_rElement[1][2] = rTemp12;
	m_rElement[2][2] = rTemp22;
	m_rElement[3][2] = rTemp32;

	m_rElement[0][3] = rTemp03;
	m_rElement[1][3] = rTemp13;
	m_rElement[2][3] = rTemp23;
	m_rElement[3][3] = rTemp33;

#if defined(_ASSERT_ADD_TRANSLATION) && defined(_DEBUG)
	{
		CMatrix	mtrxTranslation,
				mtrxResult;
		mtrxTranslation.SetTranslation(x, y, z);
		mtrxResult = mtrxOrg * mtrxTranslation;

		for(int i=0; i<4; i++){
			for(int j=0; j<4; j++){
				assert(m_rElement[j][i]==mtrxResult.m_rElement[j][i]);
			}
		}
	}
#endif

	return *this;
}

CMatrix& CMatrix::SetTranslation(double x, double y, double z)
{
	m_rElement[0][0] = double(1.0);
	m_rElement[1][0] = double(0.0);
	m_rElement[2][0] = double(0.0);
	m_rElement[3][0] = double(0.0);

	m_rElement[0][1] = double(0.0);
	m_rElement[1][1] = double(1.0);
	m_rElement[2][1] = double(0.0);
	m_rElement[3][1] = double(0.0);

	m_rElement[0][2] = double(0.0);
	m_rElement[1][2] = double(0.0);
	m_rElement[2][2] = double(1.0);
	m_rElement[3][2] = double(0.0);

	m_rElement[0][3] = x;
	m_rElement[1][3] = y;
	m_rElement[2][3] = z;
	m_rElement[3][3] = double(1.0);

	return *this;
}

CMatrix& CMatrix::SetRotation(double x, double y, double z, double rot)
{
	double	nSin = double(sin(rot)),
			nCos = double(cos(rot)),
			nSinX = x*nSin,
			nSinY = y*nSin,
			nSinZ = z*nSin,
			nCos2 = 1-nCos,
			x2 = x*x,
			y2 = y*y,
			z2 = z*z,
			xy = x*y,
			yz = y*z,
			xz = z*x;

	m_rElement[0][0] = x2 + (1-x2)*nCos;
	m_rElement[1][0] = xy*nCos2 + nSinZ;
	m_rElement[2][0] = xz*nCos2 - nSinY;
	m_rElement[3][0] = double(0.0);

	m_rElement[0][1] = xy*nCos2 - nSinZ;
	m_rElement[1][1] = y2 + (1-y2)*nCos;
	m_rElement[2][1] = yz*nCos2 + nSinX;
	m_rElement[3][1] = double(0.0);

	m_rElement[0][2] = yz*nCos2 + nSinY;
	m_rElement[1][2] = yz*nCos2 - nSinX;
	m_rElement[2][2] = z2 + (1-z2)*nCos;
	m_rElement[3][2] = double(0.0);

	m_rElement[0][3] = double(0.0);
	m_rElement[1][3] = double(0.0);
	m_rElement[2][3] = double(0.0);
	m_rElement[3][3] = double(1.0);

	return *this;
}

CMatrix& CMatrix::AddScale(double x, double y, double z)
{
#if defined(_ASSERT_ADD_SCALE) && defined(_DEBUG)
	CMatrix	mtrxOrg = *this;
#endif

	double	rTemp00 = m_rElement[0][0] * x,
			rTemp10 = m_rElement[1][0] * y,
			rTemp20 = m_rElement[2][0] * z,
//			rTemp30 = m_rElement[3][0],
			
			rTemp01 = m_rElement[0][1] * x,
			rTemp11 = m_rElement[1][1] * y,
			rTemp21 = m_rElement[2][1] * z,
//			rTemp31 = m_rElement[3][1],

			rTemp02 = m_rElement[0][2] * x,
			rTemp12 = m_rElement[1][2] * y,
			rTemp22 = m_rElement[2][2] * z,
//			rTemp32 = m_rElement[3][2],

			rTemp03 = m_rElement[0][3] * x,
			rTemp13 = m_rElement[1][3] * y,
			rTemp23 = m_rElement[2][3] * z;
//			rTemp33 = m_rElement[3][3];

	m_rElement[0][0] = rTemp00;
	m_rElement[1][0] = rTemp10;
	m_rElement[2][0] = rTemp20;
//	m_rElement[3][0] = rTemp30;

	m_rElement[0][1] = rTemp01;
	m_rElement[1][1] = rTemp11;
	m_rElement[2][1] = rTemp21;
//	m_rElement[3][1] = rTemp31;

	m_rElement[0][2] = rTemp02;
	m_rElement[1][2] = rTemp12;
	m_rElement[2][2] = rTemp22;
//	m_rElement[3][2] = rTemp32;

	m_rElement[0][3] = rTemp03;
	m_rElement[1][3] = rTemp13;
	m_rElement[2][3] = rTemp23;
//	m_rElement[3][3] = rTemp33;

#if defined(_ASSERT_ADD_SCALE) && defined(_DEBUG)
	{
		CMatrix	mtrxScale,
				mtrxResult;
		mtrxScale.SetScale(x, y, z);
		mtrxResult = mtrxOrg * mtrxScale;

		for(int i=0; i<4; i++){
			for(int j=0; j<4; j++){
				assert(m_rElement[j][i]==mtrxResult.m_rElement[j][i]);
			}
		}
	}		
#endif

	return *this;
}

CMatrix& CMatrix::SetScale(double x, double y, double z)
{
	m_rElement[0][0] = x;
	m_rElement[1][0] = double(0.0);
	m_rElement[2][0] = double(0.0);
	m_rElement[3][0] = double(0.0);

	m_rElement[0][1] = double(0.0);
	m_rElement[1][1] = y;
	m_rElement[2][1] = double(0.0);
	m_rElement[3][1] = double(0.0);

	m_rElement[0][2] = double(0.0);
	m_rElement[1][2] = double(0.0);
	m_rElement[2][2] = z;
	m_rElement[3][2] = double(0.0);

	m_rElement[0][3] = double(0.0);
	m_rElement[1][3] = double(0.0);
	m_rElement[2][3] = double(0.0);
	m_rElement[3][3] = double(1.0);

	return *this;
}

CMatrix& CMatrix::InverseMatrix()
{
	CMatrix	mtrx;

	for(int col=0; col<3; col++){
		for(int row=0; row<3; row++){
			mtrx.m_rElement[row][col]  = m_rElement[col][row];
			mtrx.m_rElement[col][  3] -= m_rElement[row][col] * m_rElement[row][  3];
		}
	}

#if defined(_ASSERT_INVERSE_MATRIX2) && defined(_DEBUG)
	{
		CMatrix	mtrxTemp;

		for(int col=0; col<3; col++){
			for(int row=0; row<3; row++){
				mtrxTemp.m_rElement[row][col]  = m_rElement[col][row];
			}
		}
		
		for(col=0; col<3; col++){
			double	val = double(0.0);
			for(int row=0; row<3; row++){
				mtrxTemp.m_rElement[row][col]  = m_rElement[col][row];
			}
			mtrx.m_rElement[col][3] = -val;
		}

		assert(mtrx==mtrxTemp);
	}
#endif

#if defined(_ASSERT_INVERSE_MATRIX) && defined(_DEBUG)
	{
		CMatrix	mtrxResult,
				mtrxUnit;
		mtrxResult = (*this) * mtrx;
		
		for(int col=0; col<4; col++){
			for(int row=0; row<4; row++){
				assert(mtrxResult.m_rElement[row][col]==mtrxUnit.m_rElement[row][col]);
			}
		}
	}
#endif

	return (*this = mtrx);
}

CMatrix& CMatrix::operator=(CMatrix const& mtrx)
{
	memcpy((void*)m_rElement, (void*)mtrx.m_rElement, sizeof m_rElement);
	return *this;
}

BOOL CMatrix::operator==(CMatrix const& mtrx) const
{
	for(int col=0; col<4; col++){
		for(int row=0; row<4; row++){
			if(m_rElement[row][col]!=mtrx.m_rElement[row][col]){
				return FALSE;
			}
		}
	}
	return TRUE;
}

BOOL CMatrix::operator!=(CMatrix const& mtrx) const
{
	return !((*this)==mtrx);
}

CMatrix CMatrix::operator*(CMatrix const& mtrx) const
{
	CMatrix	mtrxResult;

	for(int i=0; i<4; i++){
		for(int j=0; j<4; j++){
			double	val = double(0.0);
			
			for(int k=0; k<4; k++){
				val +=
					m_rElement[k][j] *
					mtrx.m_rElement[i][k];
			}

			mtrxResult.m_rElement[i][j] = val;
		}
	}
	
	return mtrxResult;
}

